# $Id: ws-mtdb.py,v 1.13 2007/09/01 06:11:15 asc Exp $

import sys
sys.path.insert(0, "/home/asc/lib/python/mtdb")

from urlparse import urlparse
from cgi import parse_qs, parse_header, parse_multipart, escape
import signal, thread, threading, time, sys
import BaseHTTPServer, SocketServer, mimetypes
import socket
import errno
import re
import textwrap

from types import *
import sqlite3
import simplejson
import mtdb

# ##############################################################

global done, server

done = False
server = None
pathdb = None

# ##############################################################

class WebServer(SocketServer.ThreadingMixIn, BaseHTTPServer.HTTPServer):
    pass

# ##############################################################

class WebRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    # ##########################################################
    
    def __init__ (self, request, client_address, server) :
        
        BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, request, client_address, server)
         
    # ##########################################################

    def do_setup (self) :

        try :
            dbconn = sqlite3.connect(pathdb)
            return mtdb.mtdb(dbconn, 0)
        except Exception, e :
            self.output_error(100, e)
            return False
    
    # ##########################################################
    
    def do_POST (self) :

        db = self.do_setup()
        
        if not db :
            return

        try :
            ctype, pdict = parse_header(self.headers.getheader('content-type'))
            length = int(self.headers.getheader('content-length'))
            
            if ctype == 'multipart/form-data':
                params = parse_multipart(self.rfile, pdict)
            else :
                qs = self.rfile.read(length)
                params = parse_qs(qs, keep_blank_values=1)
        except Exception, e :
            return self.output_error(100, e)

        params = self.validate_params(params)
        
        if not params :
            return None

        method = params['method']
    
        if method == 'mt.document.add' :
            return self.do_add_document(db, params)

        elif method == 'mt.document.remove' :
            return self.do_remove_document(db, params)
        
        else :
            return self.output_error(101, 'Unknown method')

    # ##########################################################
    
    def do_GET (self):

        db = self.do_setup()
        
        if not db :
            return
        
        scheme, host, path, params, query, hash = urlparse(self.path)

        params = parse_qs(query)
        params = self.validate_params(params)

        if not params :
            return None

        method = params['method']

        if method == 'mt.namespaces.list' :
            return self.do_introspection(db, params, 'namespace')
        
        elif method == 'mt.predicates.list' :
            return self.do_introspection(db, params, 'predicate')
        
        elif method == 'mt.values.list' :
            return self.do_introspection(db, params, 'value')
        
        elif method == 'mt.documents.search' :
            return self.do_document_search(db, params)

        elif method == 'mt.documents.search_by_range' :
            return self.do_document_search_by_range(db, params)
                
        elif method == 'mt.document.tags' :
            return self.do_tag_search(db, params)
        
        else :
            return self.output_error(101, 'Unknown method')

    # ##########################################################

    def do_introspection (self, db, params, type) :
        
        tags = []

        if type == 'namespace' :
            tags = db.find_namespaces(params['predicate'], params['value'])
        elif type == 'predicate' :
            tags = db.find_predicates(params['namespace'], params['value'])
        else :
            tags = db.find_values(params['namespace'], params['predicate'])

        plural = type + "s"
        
        out = {plural:{type:[]}}
        
        for t in tags :
            out[plural][type].append({'_content':self.sanitize(t)})
            
        out[plural]['_attrs'] = {'count':len(out[plural][type])}
        
        return self.output_ok(out, params['format'])
    
    # ##########################################################
    
    def do_document_search (self, db, params) :

        try :
            res = db.find_documents(params['namespace'], params['predicate'], params['value'])
        except Exception, e :
            return self.output_error(300, e)

        return self.__output_search(res, params)

    # ##########################################################

    def do_document_search_by_range (self, db, params) :

        values = params['value'].split(",")

        if len(values) < 2 :
            return self.output_error(201, 'Invalid range parameter')

        range = [ float(values[0]), float(values[1]) ]

        try :
            res = db.find_documents_by_range(range, params['namespace'], params['predicate'])
        except Exception, e :
            return self.output_error(300, e)

        return self.__output_search(res, params)

    # ##########################################################
    
    def __output_search (self, res, params) :
        
        out = {'documents':{'document':[]}}

        if res :
            for d in res :

                data = {'_content' : self.sanitize(d)}
                out['documents']['document'].append(data)

        out['documents']['_attrs'] = {'count':len(out['documents']['document'])}
        
        return self.output_ok(out, params['format'])
    
    # ##########################################################

    def do_tag_search (self, db, params) :

        if not params.has_key('uri') :
            return self.output_error(200, 'Missing parameter : uri')
        
        try :
            res = db.find_tags(params['uri'])
        except Exception, e :
            return self.output_error(300, e)
        
        out = {'tags':{'tag':[]}}

        if res :
            for mt in res :
                out['tags']['tag'].append({'_content':self.sanitize(mt.as_string())})

        out['tags']['_attrs'] = {'count':len(out['tags']['tag'])}
        
        return self.output_ok(out, params['format'])
        
    # ##########################################################

    def do_add_document (self, db, params) :

        if not params.has_key('uri') :
            return self.output_error(200, 'Missing parameter : uri')

        if not params.has_key('tags') :
            return self.output_error(200, 'Missing parameter : tags')

        if len(params['tags']) == 0 :
            return self.output_error(201, 'Invalid parameter : tags')
        
        try :
            db.add_document(params['uri'], params['tags'])
        except Exception, e :
            return self.output_error(300, e)

        return self.output_ok(None, params['format'])
    
    # ##########################################################

    def do_remove_document (self, db, params) :

        if not params.has_key('uri') :
            return self.output_error(200, 'Missing parameter : uri')

        try :
            db.remove_document(params['uri'])
        except Exception, e :
            return self.output_error(300, e)

        return self.output_ok(None, params['format'])
        
    # ##########################################################

    def send_output(self, status, ctype, output, extra=None) :

        try :
            self.send_response(status, "OK")
            self.send_header("Content-Type", ctype)
            self.send_header('Content-length', len(output)-1)

            if extra :
                for header in extra :
                    self.send_header(header, extra[header])

            self.send_header("Connection", "close")                    
            self.end_headers()
            
            self.wfile.write(output)
            return
        
        except socket.error, (code, msg):

            # grrr....            
            # http://trac.edgewall.org/ticket/1183
            
            if code == errno.EPIPE:
                self.log_error('Lost connection to client: %s', self.address_string())
            else:
                raise Exception (code, msg)
            
        except Exception, e :
            raise Exception(e)

    # ##########################################################
    
    def output_ok (self, data=None, format='xml') :

        ctype = self.output_content_type(format)
        output = self.output_string(data, format)
        return self.send_output(200, ctype, output)

    # ##########################################################

    def output_error (self, err_code=999, err_msg="OH NOES!!! INVISIBLE ERRORZ!!!", format='xml') :

        err_code = self.sanitize(err_code)
        err_msg  = self.sanitize(err_msg)

        data = {'error': {'_attrs' : {'code':err_code}, '_content':err_msg}}
        output = self.output_string(data, format)

        ctype = self.output_content_type(format)

        extra = {'X-ErrorCode': err_code, 'X-ErrorMessage': err_msg}

        return self.send_output(500, ctype, output, extra)

    # ##########################################################

    def output_string (self, data, format) :

        if not data :
            data = {'_attrs':{'stat':'ok'}}
            
        output = ''
        
        if format == 'json' :
            return simplejson.dumps(data)
        else :
            el = data.keys()[0]
            return self.output_xml(el, data[el])

    # ##########################################################
    
    def output_content_type (self, format='xml') :

        if format == 'xml' :
            return 'text/xml'

        return 'text/plain'

    # ##########################################################

    def output_xml (self, el, data) :

        dtype = type(data)
        xml = ""

        cl_el = self.sanitize(el)

        if dtype == ListType :

            for item in data :
                xml += self.output_xml(el, item)
                
        elif dtype == DictType :

            xml += "<%s" % cl_el

            if data.has_key('_attrs') :

                for a in data['_attrs'] :
                    cl_a = self.sanitize(a)
                    cl_v = self.sanitize(data['_attrs'][a])
                    xml +=" %s=\"%s\"" % (cl_a, cl_v)
                    
                del(data['_attrs'])
                
            xml += ">"

            if data.has_key('_content') :
                xml += self.sanitize(data['_content'])
            else :
                for i in data.keys() :
                    xml += self.output_xml(i, data[i])                

            xml += "</%s>" % cl_el
            
        else :
            xml += "<!-- WOT IS %s ??? -->" % self.sanitize(dtype)

        return xml
    
    # ##########################################################    
    
    def validate_params (self, params) :

        if len(params.keys()) == 0 :
            self.help()
            return False

        #
        
        if not params.has_key('method') :
            return self.output_error(200, 'Missing parameter : method')
    
        #

        ns = None
        pred = None
        value = None
        
        if params.has_key('namespace') :
            ns = params['namespace'][0]

        if params.has_key('predicate') :
            pred = params['predicate'][0]

        if params.has_key('value') :
            value = params['value'][0]

        args = {
            'method' : params['method'][0],
            'namespace' : ns,
            'predicate' : pred,
            'value' : value,
            }
        
        if params.has_key('format') :
            args['format'] = params['format'][0]
        else :
            args['format'] = 'xml'

        if params.has_key('uri') :
            args['uri'] = params['uri'][0]

        if params.has_key('tags') :
            args['tags'] = []
            
            for t in params['tags'][0].split(",") :
                args['tags'].append(t.strip())
                
        return args
    
    # ##########################################################        

    def help (self) :
        
        self.send_response(200, "OK")
        self.send_header("Content-Type", "text/plain")
        self.end_headers()

        self.wfile.write("ws-mtdb.py - a bare bones HTTP interface to the machine tag database (mtbd)\n\n")

        self.help_header("Example")
        self.help_para("http://127.0.0.1:9991/?method=mt.document.search_by_range&value=19,45")

        self.wfile.write("""<documents count="2">
	<document>http://www.flickr.com/photos/bopuc</document>
        <document>http://www.flickr.com/photos/bar</document>
</documents>

""")

        self.help_header("GET Methods")

        self.help_subheader("mt.namespaces.list")
        self.help_option("predicate", "An optional string representing a predicate used to filter the query. For example, all the values where the predicate is 'title'.", False)        
        self.help_option("value", "An optional string representing a value used to filter the query. For example, all the predicates where the value is 'barcelona'.", False)
        self.help_para("Returns a list of document URIs")
        
        self.help_subheader("mt.predicates.list")
        self.help_option("namespace", "An optional string representing a namespace used to filter the query. For example, all the predicates where the namespace is 'dc'.", False)
        self.help_option("value", "An optional string representing a value used to filter the query. For example, all the predicates where the value is 'barcelona'.", False)
        self.help_para("Returns a list of document URIs")
        
        self.help_subheader("mt.values.list")
        self.help_option("namespace", "An optional string representing a namespace used to filter the query. For example, all the values where the namespace is 'dc'.", False)
        self.help_option("predicate", "An optional string representing a predicate used to filter the query. For example, all the values where the predicate is 'title'.", False)
        self.help_para("Returns a list of document URIs")
        
        self.help_subheader("mt.documents.search")
        self.help_option("namespace", "An optional string representing a namespace used to filter the query. For example, all the documents where the namespace is 'dc'.", False)
        self.help_option("namespace", "An optional string representing a predicate used to filter the query. For example, all the documents where the predicate is 'title'.", False)
        self.help_option("value", "An optional string representing a value used to filter the query. For example, all the documents where the value is 'barcelona'.", False)
        self.help_para("Returns a list of document URIs")
        
        self.help_subheader("mt.documents.search_by_range")
        self.help_option("value", "A comma-separated list of two values to use as range constraints against the value of the machine tags associated with a document", True)        
        self.help_option("namespace", "An optional string representing a namespace used to filter the query. For example, all the documents where the namespace is 'dc'.", False)
        self.help_option("namespace", "An optional string representing a predicate used to filter the query. For example, all the documents where the predicate is 'title'.", False)
        self.help_para("Returns a list of document URIs")
                
        self.help_subheader("mt.document.tags")
        self.help_option("uri", "The URI of the document you want to list machine tags for", True)
        self.help_para("Returns a list of machine tags")

        self.help_header("POST Methods")

        self.help_subheader("mt.document.add")
        self.help_option("uri", "A URI for the document you want to add to the machinetag database", True)
        self.help_option("tags", "A comma-separated list of machine tags to associate with the document URI", True)        
        self.help_para("Returns OK on success")
        
        self.help_subheader("mt.document.remove")
        self.help_option("uri", "The URI of the document you want to remove from the machinetag database", True)
        self.help_para("Returns OK on success")

        self.help_header("Additional parameters")
        self.help_option("format", "Valid options are 'xml' and 'json'. Set this value to indicate the format you would like to receive API responses in. The default format is 'xml'", False)
        
        self.help_header("Errors")
        self.help_para("Errors are returned with the HTTP status code 500. Specific error codes and messages are returned both in the message body as XML and in the 'X-ErrorCode' and 'X-ErrorMessage' headers.")

        self.help_header("Questions")
        self.help_qa("Is it fast?", "Not really. It is designed, primarily, to be run on the same machine that is calling the interface.") 
        self.help_qa("Will it ever be fast?", "Sure. It is on The List (tm) to create a mod_python and/or wsgi version. Patches are welcome.")
        self.help_qa("Will it denormalize URLs?", "No. Why do you hate freedom? Or at least me?")
        self.help_qa("Can I get a pony?", "No.")
        self.help_qa("How about a machine pony?", "Go away.")        

        self.help_header("License")
        self.help_para("Copyright (c) 2007 Aaron Straup Cope. All Rights Reserved. This is free software. You may redistribute it and/or modify it under the same terms the Perl Artistic License.")

    def help_para(self, text) :

        self.wfile.write(textwrap.fill(text, 72))
        self.wfile.write("\n\n")
        
    def help_header(self, title) :
        ln = "-" * 72
        self.wfile.write("%s\n" % ln);
        self.wfile.write("%s\n" % title.upper())
        self.wfile.write("%s\n\n" % ln);

    def help_subheader(self, title) :
        ln = "-" * 72

        self.wfile.write(" %s\n" % title)
        self.wfile.write("%s\n\n" % ln);
        
    def help_option(self, opt, desc, required) :

        present = "required"

        if not required :
            present = "optional"
            
        self.wfile.write("* %s (%s)\n\n%s\n\n" % (opt, present, textwrap.fill(desc, 72, initial_indent="\t", subsequent_indent="\t")))
        
    def help_qa(self, question, answer) :
        self.wfile.write("%s\n\n" % question)
        self.wfile.write("%s\n\n" % textwrap.fill(answer, 72, initial_indent="\t", subsequent_indent="\t"))
        
    # ##########################################################
    
    def sanitize (self, txt) :
        return escape(unicode(txt))
    
# ##############################################################

def serve(port):    
    signal.signal(signal.SIGINT, terminate)
    thread.start_new_thread(runServer, (port,))
    
    global done
    while not done:
        try:
            time.sleep(0.3)
        except IOError:
            pass
   
    global server
    server.server_close()
    
def runServer(port):

    url = "http://127.0.0.1:%s" % port    
    print "ws-mtdb server running on port %s" % port
    print "connecting to database %s" % pathdb
    print "documentation and usage is available at %s/\n\n" % url

    warn = """DANGER WILL ROBINSON!!!
This is what we call a \"proof-of-concept\". It has no authentication mechanism.
If you put it on the public Internets and someone finds it they will be able to not
only query but delete your data"""

    print textwrap.fill(warn, 72)
    
    global server
    server = WebServer(("", port), WebRequestHandler)
    server.allow_reuse_address = True
    server.serve_forever()

def terminate(sig_num, frame):
    global done
    done = True    

# ##############################################################

if __name__ == "__main__":

    pathdb = sys.argv[1]
    port = 9991
    
    if len(sys.argv) >= 3 :
        port = int(sys.argv[2])
        
    serve(port)
